@@ -7,10 +7,12 @@
 #include <QApplication>
 #include <QKeyEvent>
 #include <QMimeData>
+#include <QClipboard>
 #include "msgpackrequest.h"
 #include "input.h"
 #include "konsole_wcwidth.h"
 #include "util.h"
+#include "version.h"
 
 namespace NeovimQt {
 
@@ -23,7 +25,8 @@ Shell::Shell(NeovimConnector *nvim, ShellOptions opts, QWidget *parent)
 	m_resizing(false),
 	m_mouse_wheel_delta_fraction(0, 0),
 	m_neovimBusy(false),
-	m_options(opts)
+	m_options(opts),
+	m_mouseEnabled(true)
 {
 	setAttribute(Qt::WA_KeyCompression, false);
 
@@ -57,9 +60,13 @@ Shell::Shell(NeovimConnector *nvim, ShellOptions opts, QWidget *parent)
 			this, &Shell::neovimError);
 	connect(m_nvim, &NeovimConnector::processExited,
 			this, &Shell::neovimExited);
+	connect(this, &ShellWidget::shellFontChanged,
+			this, &Shell::fontChanged);
 	connect(this, &ShellWidget::fontError,
 			this, &Shell::fontError);
 
+	m_nvim->setRequestHandler(new ShellRequestHandler(this));
+
 	if (m_nvim->isReady()) {
 		init();
 	}
@@ -142,8 +149,17 @@ void Shell::setAttached(bool attached)
 		if (isWindow()) {
 			updateGuiWindowState(windowState());
 		}
-		m_nvim->api0()->vim_command("runtime plugin/nvim_gui_shim.vim");
-		m_nvim->api0()->vim_command("runtime! ginit.vim");
+		auto req_shim = m_nvim->api0()->vim_command("runtime plugin/nvim_gui_shim.vim");
+		connect(req_shim, &MsgpackRequest::error, this, &Shell::handleShimError);
+		auto gviminit = qgetenv("GVIMINIT");
+		if (gviminit.isEmpty()) {
+			auto req_ginit = m_nvim->api0()->vim_command("runtime! ginit.vim");
+			connect(req_ginit, &MsgpackRequest::error, this, &Shell::handleGinitError);
+		} else {
+			m_nvim->api0()->vim_command(gviminit);
+		}
+
+		updateClientInfo();
 
 		// Noevim was not able to open urls till now. Check if we have any to open.
 		if(!m_deferredOpen.isEmpty()){
@@ -151,6 +167,15 @@ void Shell::setAttached(bool attached)
 			m_deferredOpen.clear();    //Neovim may change state. Clear to prevent reopening.
 		}
 
+		// Show v:errmsg if available
+		auto api1 = m_nvim->api1();
+		auto req = api1->nvim_get_vvar("errmsg");
+		connect(req, &MsgpackRequest::finished, [api1](quint32 m, quint64 f, const QVariant& r) {
+			auto err = r.toString();
+			if (!err.isEmpty()) {
+				api1->nvim_err_writeln(err.toLatin1());
+			}
+		});
 	}
 	emit neovimAttached(attached);
 	update();
@@ -427,9 +452,9 @@ void Shell::handleRedraw(const QByteArray& name, const QVariantList& opargs)
 	} else if (name == "set_scroll_region"){
 		handleSetScrollRegion(opargs);
 	} else if (name == "mouse_on"){
-		// See :h mouse
+		handleMouse(true);
 	} else if (name == "mouse_off"){
-		// See :h mouse
+		handleMouse(false);
 	} else if (name == "mode_change"){
 		if (opargs.size() < 1 || !opargs.at(0).canConvert<QByteArray>()) {
 			qWarning() << "Unexpected argument for change_mode:" << opargs;
@@ -498,6 +523,10 @@ void Shell::handleRedraw(const QByteArray& name, const QVariantList& opargs)
 		handlePopupMenuSelect(opargs.at(0).toLongLong());
 	} else if (name == "popupmenu_hide") {
 		m_pum.hide();
+	} else if (name == "mode_info_set") {
+		// TODO
+	} else if (name == "default_colors_set") {
+		// TODO
 	} else {
 		qDebug() << "Received unknown redraw notification" << name << opargs;
 	}
@@ -667,6 +696,39 @@ void Shell::handleNeovimNotification(const QByteArray &name, const QVariantList&
 		} else if (guiEvName == "Option" && args.size() >= 3) {
 			QString option = m_nvim->decode(args.at(1).toByteArray());
 			handleExtGuiOption(option, args.at(2));
+		} else if (guiEvName == "SetClipboard" && args.size() >= 4) {
+			QStringList lines = args.at(1).toStringList();
+			QString type = args.at(2).toString();
+			QString reg_name = args.at(3).toString();
+
+			if (reg_name != "*" && reg_name != "+") {
+				m_nvim->api0()->vim_report_error(m_nvim->encode("Cannot set register via GUI"));
+				return;
+			}
+
+			// FIXME proper newline char
+			QString data = lines.join("\n");
+
+			QByteArray payload;
+			QDataStream serialize(&payload, QIODevice::WriteOnly);
+			serialize << type;
+
+			// Store the selection type in the clipboard
+			QMimeData *clipData = new QMimeData();
+			clipData->setText(data);
+			clipData->setData("application/x-nvim-selection-type", payload);
+
+			auto clipboard = QClipboard::Clipboard;
+			if (reg_name == "*") {
+#if defined(Q_OS_MAC) || defined(Q_OS_WIN32)
+				clipboard = QClipboard::Clipboard;
+#else
+				clipboard = QClipboard::Selection;
+#endif
+			}
+
+			qDebug() << "Neovim changed clipboard" << data << type << reg_name << clipboard;
+			QGuiApplication::clipboard()->setMimeData(clipData, clipboard);
 		}
 		return;
 	} else if (name != "redraw") {
@@ -730,11 +792,24 @@ void Shell::handleSetOption(const QString& name, const QVariant& value)
 		emit neovimExtTablineSet(value.toBool());
 	} else if (name == "ext_popupmenu") {
 		emit neovimExtPopupmenuSet(value.toBool());
+	// TODO
+	} else if (name == "arabicshape") {
+	} else if (name == "ambiwidth") {
+	} else if (name == "emoji") {
+	} else if (name == "termguicolors") {
+	} else if (name == "ext_cmdline") {
+	} else if (name == "ext_wildmenu") {
 	} else {
 		qDebug() << "Received unknown option" << name << value;
 	}
 }
 
+// Enable/Disable mouse support. See mouse_off/mouse_on in :h ui-global.
+void Shell::handleMouse(bool enabled)
+{
+	m_mouseEnabled = enabled;
+}
+
 void Shell::paintEvent(QPaintEvent *ev)
 {
 	if (!m_attached) {
@@ -787,7 +862,7 @@ void Shell::keyPressEvent(QKeyEvent *ev)
 
 void Shell::neovimMouseEvent(QMouseEvent *ev)
 {
-	if (!m_attached) {
+	if (!m_attached || !m_mouseEnabled) {
 		return;
 	}
 
@@ -920,6 +995,26 @@ void Shell::updateWindowId()
 		m_nvim->connectionType() == NeovimConnector::SpawnedConnection) {
 		WId window_id = effectiveWinId();
 		m_nvim->api0()->vim_set_var("GuiWindowId", QVariant(window_id));
+		m_nvim->api0()->vim_command(QString("let v:windowid = %1").arg(window_id).toLatin1());
+		updateClientInfo();
+	}
+}
+
+void Shell::updateClientInfo()
+{
+	if (m_attached) {
+		auto api4 = m_nvim->api4();
+		if (api4) {
+			WId window_id = effectiveWinId();
+			auto version = QVariantMap();
+			version.insert("major", PROJECT_VERSION_MAJOR);
+			version.insert("minor", PROJECT_VERSION_MINOR);
+			version.insert("patch", PROJECT_VERSION_PATCH);
+			QVariantMap attrs;
+			attrs.insert("windowid", window_id);
+			attrs.insert("gui-clipboard", true);
+			api4->nvim_set_client_info("nvim-qt", version, "ui", QVariantMap(), attrs);
+		}
 	}
 }
 
@@ -1019,7 +1114,8 @@ void Shell::closeEvent(QCloseEvent *ev)
 		// If attached to a spawned Neovim process, ignore the event
 		// and try to close Neovim as :qa
 		ev->ignore();
-		m_nvim->api0()->vim_command("qa");
+		bailoutIfinputBlocking();
+		m_nvim->api0()->vim_command("confirm qa");
 	} else {
 		QWidget::closeEvent(ev);
 	}
@@ -1226,4 +1322,113 @@ void Shell::openFiles(QList<QUrl> urls)
 	}
 }
 
+// If neovim is blocked waiting for input, attempt to bailout from
+// whatever neovim is doing by pressing Ctrl-C.
+void Shell::bailoutIfinputBlocking()
+{
+	auto api = m_nvim->api2();
+	if (api) {
+		auto req = api->nvim_get_mode();
+
+		connect(req, &MsgpackRequest::finished, [api](quint32 msgid, quint64 f, const QVariant& r) {
+				auto map = r.toMap();
+				if (map.value("blocking", false) == true) {
+					api->nvim_input("<C-c>");
+				}
+		});
+	}
+}
+
+ShellRequestHandler::ShellRequestHandler(Shell *parent)
+:QObject(parent)
+{
+}
+
+const char SELECTION_MIME_TYPE[] = "application/x-nvim-selection-type";
+void ShellRequestHandler::handleRequest(MsgpackIODevice* dev, quint32 msgid, const QByteArray& method, const QVariantList& args)
+{
+	if (method == "Gui" && args.size() > 0) {
+		QString ctx = args.at(0).toString();
+		if (ctx == "GetClipboard" && args.size() > 1) {
+			QVariant reg = args.at(1);
+			QString reg_name = reg.toString();
+
+			if (reg_name != "*" && reg_name != "+") {
+				dev->sendResponse(msgid, QString("Unknown register"), QVariant());
+				return;
+			}
+
+			// + by default
+			auto mode = QClipboard::Clipboard;
+			if (reg_name == "*") {
+#if defined(Q_OS_MAC) || defined(Q_OS_WIN32)
+				mode = QClipboard::Clipboard;
+#else
+				mode = QClipboard::Selection;
+#endif
+			}
+
+			// Check nvim, ops.c/get_clipboard() - Expected to return a list with two items
+			// [register data, selection type]. The type can be ommited.
+			QVariantList result;
+
+			auto clipboard_data = QGuiApplication::clipboard()->mimeData(mode);
+			auto data = clipboard_data->text();
+			qDebug() << data << "<<<<< clipboard text";
+			// The register data is either a string with a single line,
+			// or a list of strings for multiple lines.
+			if (data.contains("\n")) {
+				result.append(data.split("\n"));
+			} else {
+				result.append(QStringList() << data);
+			}
+
+			// If available, deserialize the motion type from the clipboard
+			if (clipboard_data->hasFormat(SELECTION_MIME_TYPE)) {
+				QString type;
+				QDataStream serialize(clipboard_data->data(SELECTION_MIME_TYPE));
+				serialize >> type;
+				result.append(type);
+			} else {
+				result.append("");
+			}
+
+			qDebug() << "Neovim requested clipboard contents" << args << mode << "->" << result;
+			dev->sendResponse(msgid, QVariant(), result);
+			return;
+		}
+	}
+	// be sure to return early or this message will be sent
+	dev->sendResponse(msgid, QString("Unknown method"), QVariant());
+}
+
+/**
+ * Convert neovim error response into an error message string. If this fails
+ * serialize the entire error response as with QDebug.
+ */
+QString Shell::neovimErrorToString(const QVariant& err)
+{
+	auto lst = err.toList();
+	if (1 < lst.size()) {
+		return lst.at(1).toByteArray();
+	} else {
+		QString payload;
+		QDebug dbg(&payload);
+		dbg << err;
+		return payload;
+	}
+}
+
+void Shell::handleGinitError(quint32 msgid, quint64 fun, const QVariant& err)
+{
+	qDebug() << "ginit.vim error " << err;
+	auto msg = neovimErrorToString(err);
+	m_nvim->api0()->vim_report_error("ginit.vim error: " + msg.toUtf8());
+}
+
+void Shell::handleShimError(quint32 msgid, quint64 fun, const QVariant& err)
+{
+	qDebug() << "GUI shim error " << err;
+}
+
 } // Namespace
